#ifndef f_MPEG2DECODER_H
#define f_MPEG2DECODER_H

class IMPEG2Decoder {

public:
	enum {
		kError						= 1,	// it's just... error.
		kErrorNotInitialized		= 2,
		kErrorBadValue				= 4,
		kErrorBadFrameType			= 8,
		kErrorBadReference			= 16,
		kErrorBadMotionVector		= 32,
		kErrorTooManyCoefficients	= 64,
		kErrorTooManyMacroblocks	= 128,
		kErrorSourceOverrun			= 256
	};
	
	enum kAcceleration {
		eAccel_None	= 0,
		eAccel_MMX	= 1,
		eAccel_ISSE	= 2
	};
	
	virtual bool Init(int width, int height, long seq_ext) = 0;
	virtual void Destruct() = 0;

	// decoding
	
	virtual int DecodeFrame(const void *src, long len, long frame, int dst, int fwd, int rev) = 0;
	virtual long GetErrorState() = 0;
	virtual int GetFrameBuffer(long frame) = 0;
	virtual long GetFrameNumber(int buffer) = 0;
	virtual void CopyFrameBuffer(int dst, int src, long newframe) = 0;
	virtual void SwapFrameBuffers(int dst, int src) = 0;
	virtual void ClearFrameBuffers() = 0;

	virtual void CopyField(int dst, int dstfield, int src, int srcfield) = 0;
	virtual void EnableMatrixCoefficients(bool bEnable) = 0;
	virtual void EnableAcceleration(int level) = 0;

	// framebuffer access

	virtual const void *GetYBuffer(int buffer, long& pitch) = 0;
	virtual const void *GetCrBuffer(int buffer, long& pitch) = 0;
	virtual const void *GetCbBuffer(int buffer, long& pitch) = 0;

	// framebuffer conversion

	virtual bool DecodeUYVY(void *dst, long pitch, int buffer) = 0;
	virtual bool DecodeYUYV(void *dst, long pitch, int buffer) = 0;
	virtual bool DecodeYVYU(void *dst, long pitch, int buffer) = 0;
	virtual bool DecodeY41P(void *dst, long pitch, int buffer) = 0;
	virtual bool DecodeRGB15(void *dst, long pitch, int buffer) = 0;
	virtual bool DecodeRGB16(void *dst, long pitch, int buffer) = 0;
	virtual bool DecodeRGB24(void *dst, long pitch, int buffer) = 0;
	virtual bool DecodeRGB32(void *dst, long pitch, int buffer) = 0;

	virtual bool Decode420Planar(void *dstY, long pitchY, void *dstU,
		long pitchU, void *dstV, long pitchV, int buffer) = 0;

	virtual bool Decode422Planar(void *dstY, long pitchY, void *dstU,
		long pitchU, void *dstV, long pitchV, int buffer) = 0;
};

IMPEG2Decoder *CreateMPEG2Decoder();

#endif	// f_MPEG2DECODER_H
