///////////////////////////////////////////////////////////////////////
// MPEG-2 Plugin for VirtualDub 1.8.1+
// Copyright (C) 2007-2008 fccHandler
// Copyright (C) 1998-2008 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// Read/Write index (.midx) files

#include <windows.h>
#include <stdio.h>
#include "InputFileMPEG2.h"

#define VARIABLE_SIZES		1
#define VARIABLE_HEADERS	2
#define VARIABLE_POSITIONS	4
#define LONG_DELTAS			8
#define SHORT_DELTAS		16
#define LONG_SIZES			32
#define SHORT_SIZES			64


LPWSTR InputFileMPEG2::GetIndexName(LPCWSTR szFile) {
	int len = lstrlenW(szFile);
	if (len > 0) {

		LPWSTR ret = (LPWSTR)malloc((len + 5 + 1) * 2);
		if (ret != NULL) {
			memcpy(ret, szFile, (len + 1) * 2);

			if (len > 5) {
				const __int64 ext = *(__int64 *)(ret + len - 4);
				if ((ext | 0x0020002000200020) == 0x007800640069006D) {
					if (*(ret + len - 5) == L'.') return ret;
				}
			}

			// Append ".idx" extension
			ret[len + 0] = L'.';
			ret[len + 1] = L'm';
			ret[len + 2] = L'i';
			ret[len + 3] = L'd';
			ret[len + 4] = L'x';
			ret[len + 5] = 0;

			return ret;
		}
	}

	return NULL;
}

void InputFileMPEG2::ReadPacketInfo19(inFile& idxFile, MPEGPacketInfo *mpi, unsigned long packets) {
	unsigned long delta, i;
	int flags;

	idxFile.inRead(mpi, sizeof(MPEGPacketInfo));
	idxFile.inRead(&flags, sizeof(int));

	for (i = 1; i < packets; ++i) {
		if (flags == SHORT_DELTAS) {
			delta = 0;
			idxFile.inRead(&delta, sizeof(short));
		} else {
			idxFile.inRead(&delta, sizeof(long));
		}
		mpi[i].file_pos = ((mpi[i - 1].file_pos >> 16) + delta) << 16;

		idxFile.inRead(&(mpi[i].file_pos), sizeof(short));
	}
}

void InputFileMPEG2::WritePacketInfo(inFile& idxFile, const MPEGPacketInfo *mpi, unsigned long packets) {
	unsigned long delta, i;
	int flags;

	// Write the initial packet
	idxFile.outWrite(mpi, sizeof(MPEGPacketInfo));

	// If we can use shorts, do so, otherwise use longs
	flags = SHORT_DELTAS;
	for (i = 1; i < packets; ++i) {
		delta = (unsigned long)(
			(mpi[i].file_pos >> 16) - (mpi[i - 1].file_pos >> 16)
		);
		if (delta > 65535) {
			flags = LONG_DELTAS;
			break;
		}
	}

	idxFile.outWrite(&flags, sizeof(int));

	for (i = 1; i < packets; ++i) {
		delta = (unsigned long)(
			(mpi[i].file_pos >> 16) - (mpi[i - 1].file_pos >> 16)
		);
		if (flags == SHORT_DELTAS) {
			idxFile.outWrite(&delta, sizeof(short));
		} else {
			idxFile.outWrite(&delta, sizeof(long));
		}
		delta = (unsigned long)(mpi[i].file_pos & 0xFFFF);
		idxFile.outWrite(&delta, sizeof(short));
	}
}

void InputFileMPEG2::ReadSampleInfo19(inFile& idxFile, MPEGSampleInfo *msi, int samples) {
	for (int i = 0; i < samples; ++i) {
		idxFile.inRead(msi, sizeof(MPEGSampleInfo));
		++msi;
	}
}

void InputFileMPEG2::WriteSampleInfo(inFile& idxFile, const MPEGSampleInfo *msi, int samples) {
	for (int i = 0; i < samples; ++i) {
		idxFile.outWrite(msi, sizeof(MPEGSampleInfo));
		++msi;
	}
}

bool InputFileMPEG2::ReadIndex20(inFile& idxFile, AppendNames *an) {
	int i, nf;

	an->num_files = 0;

	idxFile.inRead(&nf, sizeof(int));
	if (nf <= 0) goto Abort;

	for (i = 0; i < nf; ++i) {
		bool success;
		int len = 0;

		idxFile.inRead(&len, sizeof(int));
		if (len <= 0 || len >= MAX_PATH) break;

		an->fName[i] = (LPWSTR)malloc(len * 2);
		if (an->fName[i] == NULL) break;

		idxFile.inRead(an->fName[i], len * 2);
		if (i == 0) {
			success = mFile.inOpen(an->fName[i]);
		} else {
			success = mFile.inAppend(an->fName[i]);
		}

		if (!success) {
			free(an->fName[i]);
			an->fName[i] = NULL;
			break;
		}

		++an->num_files;
	}

	if (i <= 0) goto Abort;

	// Now allocate the memory resources (we want
	// to give up early if there is a memory problem)

	idxFile.inRead(&i, sizeof(int));
	fInterleaved			= ((i & 1) != 0);
	progressive_sequence	= ((i & 2) != 0);

	idxFile.inRead(&vframes, sizeof(int));
	if (vframes <= 0) goto Abort;
	idxFile.inRead(&vpackets, sizeof(long));
	if (vpackets == 0) goto Abort;

	if (fInterleaved) {
		idxFile.inRead(audio_stream_list, 96);
		for (i = 0; i < 96; ++i) {
			unsigned int id = audio_stream_list[i] - 1;
			if (id < (unsigned int)48) {
				idxFile.inRead(&(apackets[id]), sizeof(long));
				idxFile.inRead(&(aframes[id]), sizeof(int));
				idxFile.inRead(&(afirstPTS[id]), sizeof(__int64));
				if (aframes[id] > 0 && apackets[id] <= 0) goto Abort;
			}
		}
	}

	// Allocate the memory
	video_packet_list = (MPEGPacketInfo *)(new char[vpackets * sizeof(MPEGPacketInfo)]);
	if (video_packet_list == NULL) goto Abort;

	if (fInterleaved) {
		for (i = 0; i < 96; ++i) {
			unsigned int id = audio_stream_list[i] - 1;
			if (id < (unsigned int)48) {
				if (apackets[id] > 0) {
					audio_packet_list[id] = (MPEGPacketInfo *)(new char[apackets[id] * sizeof(MPEGPacketInfo)]);
					if (audio_packet_list[id] == NULL) goto Abort;

					if (aframes[id] > 0) {
						audio_sample_list[id] = (MPEGSampleInfo *)(new char[aframes[id] * sizeof(MPEGSampleInfo)]);
						if (audio_sample_list[id] == NULL) goto Abort;
					}
				}
			}
		}
	}

	video_sample_list = (MPEGSampleInfo *)(new char[vframes * sizeof(MPEGSampleInfo)]);
	if (video_sample_list == NULL) goto Abort;

	// Good to go!  Now we just read in all the information...

	ReadSampleInfo19(idxFile, video_sample_list, vframes);
	ReadPacketInfo19(idxFile, video_packet_list, vpackets);

	if (fInterleaved) {
		for (i = 0; i < 96; ++i) {
			unsigned int id = audio_stream_list[i] - 1;
			if (id < (unsigned int)48) {
				if (apackets[id] > 0) {
					ReadPacketInfo19(idxFile, audio_packet_list[id], apackets[id]);
					if (aframes[id] > 0) {
						ReadSampleInfo19(idxFile, audio_sample_list[id], aframes[id]);
					}
				}
			}
		}
	}

	idxFile.inRead(&mFrameRate, sizeof(VDXFraction));
	idxFile.inRead(&vfirstPTS, sizeof(__int64));

	idxFile.inRead(&width, sizeof(int));
	idxFile.inRead(&height, sizeof(int));
	idxFile.inRead(&aspect_ratio, sizeof(int));

//	idxFile.inRead(&iChosenAudio, sizeof(int));
	idxFile.inRead(&seq_ext, sizeof(long));

	return true;

Abort:
	return false;
}

void InputFileMPEG2::WriteIndex(LPCWSTR szFile, AppendNames *an) {
	inFile idxFile;
	DWORD buf;
	int i;

	if (an->num_files <= 0) return;

	// Create index file
	{
		bool success;
		LPWSTR idxName = GetIndexName(szFile);
		if (idxName == NULL) goto Abort;
		success = idxFile.outOpen(idxName);
		free(idxName);

		if (!success) goto Abort;
	}

	buf = 0x5844494D;	// 'MIDX'
	idxFile.outWrite(&buf, 4);
	buf = 0x00000200;	// Version 2.0
	idxFile.outWrite(&buf, 4);

	idxFile.outWrite(&(an->num_files), sizeof(int));
	for (i = 0; i < an->num_files; ++i) {
		int len = lstrlenW(an->fName[i]) + 1;
		if (len <= 0 || len >= MAX_PATH) len = 0;
		idxFile.outWrite(&len, sizeof(int));
		if (len > 0) idxFile.outWrite(an->fName[i], len * 2);
	}

	i = 0;
	if (fInterleaved)			i |= 1;
	if (progressive_sequence)	i |= 2;
	idxFile.outWrite(&i, sizeof(int));

	idxFile.outWrite(&vframes, sizeof(int));
	idxFile.outWrite(&vpackets, sizeof(long));
	if (fInterleaved) {
		idxFile.outWrite(audio_stream_list, 96);
		for (i = 0; i < 96; ++i) {
			unsigned int id = audio_stream_list[i] - 1;
			if (id < (unsigned int)48) {
				idxFile.outWrite(&(apackets[id]), sizeof(long));
				idxFile.outWrite(&(aframes[id]), sizeof(int));
				idxFile.outWrite(&(afirstPTS[id]), sizeof(__int64));
			}
		}
	}

	// Now write the arrays as applicable
	WriteSampleInfo(idxFile, video_sample_list, vframes);
	WritePacketInfo(idxFile, video_packet_list, vpackets);
	if (fInterleaved) {
		for (i = 0; i < 96; ++i) {
			unsigned int id = audio_stream_list[i] - 1;
			if (id < (unsigned int)48) {
				if (apackets[id] > 0) {
					WritePacketInfo(idxFile, audio_packet_list[id], apackets[id]);
					if (aframes[id] > 0) {
						WriteSampleInfo(idxFile, audio_sample_list[id], aframes[id]);
					}
				}
			}
		}
	}

	idxFile.outWrite(&mFrameRate, sizeof(VDXFraction));
	idxFile.outWrite(&vfirstPTS, sizeof(__int64));

	idxFile.outWrite(&width, sizeof(int));
	idxFile.outWrite(&height, sizeof(int));
	idxFile.outWrite(&aspect_ratio, sizeof(int));

//	idxFile.outWrite(&iChosenAudio, sizeof(int));
	idxFile.outWrite(&seq_ext, sizeof(long));

Abort:
	return;
}



bool InputFileMPEG2::ReadIndex(LPCWSTR szFile, AppendNames *an) {
	inFile idxFile;

	// Open the index file, if it exists
	LPWSTR idxName = GetIndexName(szFile);
	if (idxName != NULL) {
		bool success = idxFile.inOpen(idxName);
		free(idxName);

		if (success) {
			DWORD buf;
			idxFile.inRead(&buf, 4);
			if (buf == 0x5844494D) {	// 'MIDX'
				int i;
				idxFile.inRead(&buf, 4);

				switch (buf) {
					case 0x00000200:	// Version 2.0
						if (ReadIndex20(idxFile, an)) return true;
						break;
				}

				// If ReadIndex## fails, destroy everything
				delete [] video_sample_list;
				video_sample_list = NULL;
				vframes = 0;

				for (i = 47; i >= 0; --i) {
					delete [] audio_sample_list[i];
					audio_sample_list[i] = NULL;
					aframes[i] = 0;

					delete [] audio_packet_list[i];
					audio_packet_list[i] = NULL;
					apackets[i] = 0;
				}

				delete [] video_packet_list;
				video_packet_list = NULL;
				vpackets = 0;
			}
		}
	}

	return false;
}
