///////////////////////////////////////////////////////////////////////
// MPEG-2 Plugin for VirtualDub 1.8.1+
// Copyright (C) 2007-2008 fccHandler
// Copyright (C) 1998-2008 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// MPEG-2 video source (IVDXVideoSource)

#include <windows.h>

#include "decode.h"

#include <crtdbg.h>

///////////////////////////////////////////////////////////////////////////////

class VideoSourceMPEG2 : public vdxunknown<IVDXStreamSource>, public IVDXVideoSource {

private:
	InputFileMPEG2 *const parentPtr;

	unsigned char *read_buffer1;
	long buffer1_frame;
	unsigned char *read_buffer2;
	long buffer2_frame;

	sint64 mSampleFirst;
	sint64 mSampleLast;

public:
	VideoSourceMPEG2(InputFileMPEG2 *const pp);
	~VideoSourceMPEG2();

	int VDXAPIENTRY AddRef();
	int VDXAPIENTRY Release();
	void *VDXAPIENTRY AsInterface(uint32 iid);

	void		VDXAPIENTRY GetStreamSourceInfo(VDXStreamSourceInfo&);
	bool		VDXAPIENTRY Read(sint64 lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead);

	const void *VDXAPIENTRY GetDirectFormat();
	int			VDXAPIENTRY GetDirectFormatLen();

	ErrorMode VDXAPIENTRY GetDecodeErrorMode();
	void VDXAPIENTRY SetDecodeErrorMode(ErrorMode mode);
	bool VDXAPIENTRY IsDecodeErrorModeSupported(ErrorMode mode);

	bool VDXAPIENTRY IsVBR();
	sint64 VDXAPIENTRY TimeToPositionVBR(sint64 us);
	sint64 VDXAPIENTRY PositionToTimeVBR(sint64 samples);

	void VDXAPIENTRY GetVideoSourceInfo(VDXVideoSourceInfo& info);

	bool VDXAPIENTRY CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel);
	bool VDXAPIENTRY CreateVideoDecoder(IVDXVideoDecoder **ppDecoder);

	void		VDXAPIENTRY GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo);

	bool		VDXAPIENTRY IsKey(sint64 lSample);

	sint64		VDXAPIENTRY GetFrameNumberForSample(sint64 sample_num);
	sint64		VDXAPIENTRY GetSampleNumberForFrame(sint64 display_num);
	sint64		VDXAPIENTRY GetRealFrame(sint64 display_num);

	sint64		VDXAPIENTRY GetSampleBytePosition(sint64 sample_num);
};

VideoSourceMPEG2::VideoSourceMPEG2(InputFileMPEG2 *const pp)
	: parentPtr(pp)
{
	parentPtr->AddRef();

_RPT0(_CRT_WARN, "VideoSourceMPEG2 constructed\n");

	read_buffer2 = NULL;
	read_buffer1 = (unsigned char *)malloc(parentPtr->largest_framesize * 2);
	if (read_buffer1) read_buffer2 = read_buffer1 + parentPtr->largest_framesize;

	buffer1_frame = -1;
	buffer2_frame = -1;

	mSampleFirst = 0;
	mSampleLast = parentPtr->fields;
}

VideoSourceMPEG2::~VideoSourceMPEG2() {
_RPT0(_CRT_WARN, "VideoSourceMPEG2 destructed\n");
	free(read_buffer1);
	parentPtr->Release();
}

int VideoSourceMPEG2::AddRef() {
	return vdxunknown<IVDXStreamSource>::AddRef();
}

int VideoSourceMPEG2::Release() {
	return vdxunknown<IVDXStreamSource>::Release();
}

void *VDXAPIENTRY VideoSourceMPEG2::AsInterface(uint32 iid) {
	if (iid == IVDXVideoSource::kIID)
		return static_cast<IVDXVideoSource *>(this);

	if (iid == IVDXStreamSource::kIID)
		return static_cast<IVDXStreamSource *>(this);

	return NULL;

//	return vdxunknown<IVDXStreamSource>::AsInterface(iid);
}

void VDXAPIENTRY VideoSourceMPEG2::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo) {
	srcInfo.mSampleRate = parentPtr->mFrameRate;
	srcInfo.mSampleCount = parentPtr->fields;

	// TODO: If any future build of VirtualDub starts using the
	// PAR information, figure out the correct values to return.

	switch (parentPtr->aspect_ratio) {

		// MPEG-1 SAR
		case 1:		// (1.0000) VGA
		case 2:		// (0.6735)
		case 3:		// (0.7031) 16:9 PAL
		case 4:		// (0.7615)
		case 5:		// (0.8055)
		case 6:		// (0.8437) 16:9 NTSC
		case 7:		// (0.8935)
		case 8:		// (0.9157) 4:3 PAL
		case 9:		// (0.9815)
		case 10:	// (1.0255)
		case 11:	// (1.0695)
		case 12:	// (1.0950) 4:3 NTSC
		case 13:	// (1.1575)
		case 14:	// (1.2015)

		// MPEG-2 SAR/DAR
		case 33:	// (1.0) SAR
		case 34:	// (4:3) DAR
		case 35:	// (16:9) DAR
		case 36:	// (2.21:1) DAR

		default:
			srcInfo.mPixelAspectRatio.mNumerator = 1;
			srcInfo.mPixelAspectRatio.mDenominator = 1;
			break;
	};
}

bool VideoSourceMPEG2::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {
	const MPEGSampleInfo *m1, *m2;
	long lStart, f1, f2, len;
	int cadence;

//_RPT1(_CRT_WARN, "Read(start = %i)\n", (int)lStart64);

	if (lStart64 < mSampleFirst || lStart64 >= mSampleLast) {
		if (lSamplesRead) *lSamplesRead = 0;
		if (lBytesRead) *lBytesRead = 0;
		return true;
	}

	lStart = (long)lStart64;

	f1 = parentPtr->video_field_map[lStart];
	cadence = f1 & 1;
	f1 >>= 1;

	if (cadence != 0 && lStart64 > mSampleFirst) {
		f2 = f1;
		f1 = parentPtr->video_field_map[lStart - 1] >> 1;

		m1 = &(parentPtr->video_sample_list[f1]);
		m2 = &(parentPtr->video_sample_list[f2]);

		len = m1->size + m2->size;

	} else {

		m1 = &(parentPtr->video_sample_list[f1]);
		len = m1->size;
	}

	// We only support reading one frame
	// at a time; "lCount" is never used.

	if (lpBuffer == NULL || cbBuffer < (uint32)len) {
		if (lSamplesRead) *lSamplesRead = 1;
		if (lBytesRead) *lBytesRead = len;
		return (lpBuffer == NULL);
	}

	if (cadence) {
		if (buffer1_frame == f1) {
			memcpy(lpBuffer, read_buffer1, m1->size);
			if (buffer2_frame != f2) {
				parentPtr->ReadSample(read_buffer2, f2, m2->size, 0);
				buffer2_frame = f2;
			}
			memcpy((char *)lpBuffer + m1->size, read_buffer2, m2->size);

		} else if (buffer2_frame == f1) {
			memcpy(lpBuffer, read_buffer2, m1->size);
			if (buffer1_frame != f2) {
				parentPtr->ReadSample(read_buffer1, f2, m2->size, 0);
				buffer1_frame = f2;
			}
			memcpy((char *)lpBuffer + m1->size, read_buffer1, m2->size);

		} else if (buffer1_frame == f2) {
			parentPtr->ReadSample(read_buffer2, f1, m1->size, 0);
			memcpy(lpBuffer, read_buffer2, m1->size);
			buffer2_frame = f1;
			memcpy((char *)lpBuffer + m1->size, read_buffer1, m2->size);

		} else if (buffer2_frame == f2) {
			parentPtr->ReadSample(read_buffer1, f1, m1->size, 0);
			memcpy(lpBuffer, read_buffer1, m1->size);
			buffer1_frame = f1;
			memcpy((char *)lpBuffer + m1->size, read_buffer2, m2->size);

		} else {
			parentPtr->ReadSample(read_buffer1, f1, m1->size, 0);
			parentPtr->ReadSample(read_buffer2, f2, m2->size, 0);
			buffer1_frame = f1;
			buffer2_frame = f2;
			memcpy(lpBuffer, read_buffer1, m1->size);
			memcpy((char *)lpBuffer + m1->size, read_buffer2, m2->size);
		}

	} else {
		if (buffer1_frame == f1) {
			memcpy(lpBuffer, read_buffer1, m1->size);
		} else if (buffer2_frame == f1) {
			memcpy(lpBuffer, read_buffer2, m1->size);
		} else {
			parentPtr->ReadSample(read_buffer2, f1, m1->size, 0);
			memcpy(lpBuffer, read_buffer2, m1->size);
			buffer2_frame = f1;
		}
	}

	if (lSamplesRead) *lSamplesRead = 1;
	if (lBytesRead) *lBytesRead = len;
	return true;
}

const void *VideoSourceMPEG2::GetDirectFormat() {
	return NULL;
}

int VideoSourceMPEG2::GetDirectFormatLen() {
	return 0;
}

IVDXStreamSource::ErrorMode VideoSourceMPEG2::GetDecodeErrorMode() {
	return IVDXStreamSource::kErrorModeReportAll;
}

void VideoSourceMPEG2::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode) {
}

bool VideoSourceMPEG2::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode) {
	return mode == IVDXStreamSource::kErrorModeReportAll;
}

bool VideoSourceMPEG2::IsVBR() {
	return false;
}

	// sample = (ms * dwRate) / (dwScale * 1000)
	// sample * dwScale * 1000 = (ms * dwRate)
	// ms = (sample * dwScale * 1000) / dwRate

sint64 VideoSourceMPEG2::TimeToPositionVBR(sint64 us) {
	return (sint64)(0.5 +
		((double)parentPtr->mFrameRate.mNumerator * (double)us)
		/ ((double)parentPtr->mFrameRate.mDenominator * 1000000.0)
	);
}

sint64 VideoSourceMPEG2::PositionToTimeVBR(sint64 samples) {
	return (sint64)(0.5 +
		((double)parentPtr->mFrameRate.mDenominator * (double)samples * 1000000.0)
		/ (double)parentPtr->mFrameRate.mNumerator
	);
}

void VideoSourceMPEG2::GetVideoSourceInfo(VDXVideoSourceInfo& info) {
	info.mFlags = VDXVideoSourceInfo::kFlagNone;
	info.mWidth = parentPtr->width;
	info.mHeight = parentPtr->height;
	info.mDecoderModel = VDXVideoSourceInfo::kDecoderModelCustom;
}

bool VideoSourceMPEG2::CreateVideoDecoderModel(IVDXVideoDecoderModel **ppModel) {
//	IVDXVideoDecoderModel *p = new VideoDecoderModelMPEG2(parentPtr);
	IVDXVideoDecoderModel *p = CreateVideoDecoderModelMPEG2(parentPtr);
	if (!p) return false;

	p->AddRef();
	*ppModel = p;
	return true;
}

bool VideoSourceMPEG2::CreateVideoDecoder(IVDXVideoDecoder **ppDecoder) {
//	IVDXVideoDecoder *p = new VideoDecoderMPEG2(parentPtr);
	IVDXVideoDecoder *p = CreateVideoDecoderMPEG2(parentPtr);
	if (!p)	return false;

	p->AddRef();
	*ppDecoder = p;
	return true;
}

void VideoSourceMPEG2::GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo) {
	frameInfo.mBytePosition = -1;
	frameInfo.mFrameType = kVDXVFT_Null;
	frameInfo.mTypeChar = ' ';

	if (sample_num >= mSampleFirst && sample_num < mSampleLast) {
		long f1 = parentPtr->video_field_map[(uint32)sample_num];

		if (f1 & 1) {
			frameInfo.mTypeChar = 'R';	// reconstructed
		} else {
			switch (parentPtr->video_sample_list[f1 >> 1].frame_type) {
				case MPEG_FRAME_TYPE_I:
					frameInfo.mFrameType = kVDXVFT_Independent;
					frameInfo.mTypeChar = 'I';
					break;
				case MPEG_FRAME_TYPE_P:
					frameInfo.mFrameType = kVDXVFT_Predicted;
					frameInfo.mTypeChar = 'P';
					break;
				case MPEG_FRAME_TYPE_B:
					frameInfo.mFrameType = kVDXVFT_Bidirectional;
					frameInfo.mTypeChar = 'B';
					break;
			}
		}
	}
}

bool VideoSourceMPEG2::IsKey(sint64 sample) {
	if (sample >= mSampleFirst && sample < mSampleLast) {
		long f1 = parentPtr->video_field_map[(uint32)sample];

		if ((f1 & 1) == 0) {
			if (parentPtr->video_sample_list[f1 >> 1].frame_type == MPEG_FRAME_TYPE_I) {
				return true;
			}
		}
	}

	return false;
}

sint64 VideoSourceMPEG2::GetFrameNumberForSample(sint64 sample_num) {
	return sample_num;
}

sint64 VideoSourceMPEG2::GetSampleNumberForFrame(sint64 display_num) {
	return display_num;
}

sint64 VideoSourceMPEG2::GetRealFrame(sint64 display_num) {
	return display_num;
}

sint64 VideoSourceMPEG2::GetSampleBytePosition(sint64 sample_num) {
	return -1;
}


IVDXVideoSource *CreateVideoSourceMPEG2(InputFileMPEG2 *const pp) {
	return new VideoSourceMPEG2(pp);
}